/*
 * Performance-monitoring counters driver.
 * Top-level initialisation code and implementation of /dev/perfctr.
 *
 * Copyright (C) 1999-2000  Mikael Pettersson
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/miscdevice.h>
#include <linux/smp.h>

#include <asm/uaccess.h>
#include <asm/perfctr.h>

#include "compat.h"
#include "x86.h"
#include "virtual.h"
#include "global.h"

#define PERFCTR_VERSION_MAJOR	0
#define PERFCTR_VERSION_MINOR	10
#define PERFCTR_VERSION_MICRO	0

static struct perfctr_info perfctr_info;

static int sys_perfctr_info(void *arg)
{
	if( copy_to_user(arg, &perfctr_info, sizeof perfctr_info) )
		return -EFAULT;
	return 0;
}

static int dev_perfctr_ioctl(struct inode *inode, struct file *filp,
			     unsigned int cmd, unsigned long arg)
{
	switch( cmd ) {
	case PERFCTR_INFO:
		return sys_perfctr_info((void*)arg);
	case PERFCTR_ATTACH_RDONLY:
		return vperfctr_attach((int)arg, 0);
	case PERFCTR_ATTACH_RDWR:
		return vperfctr_attach((int)arg, 1);
	case PERFCTR_VIRTUAL_SAMPLE:
		return vperfctr_sample_current();
	case PERFCTR_VIRTUAL_UNLINK:
		return vperfctr_unlink_current();
	case PERFCTR_GLOBAL_CONTROL:
		return gperfctr_control((struct gperfctr_control*)arg);
	case PERFCTR_GLOBAL_READ:
		return gperfctr_read((struct gperfctr_state*)arg);
	case PERFCTR_GLOBAL_STOP:
		return gperfctr_stop();
	}
	printk(KERN_ERR __FUNCTION__ ": bad cmd 0x%08x\n", cmd);
	return -EINVAL;
}

static int dev_perfctr_open(struct inode *inode, struct file *filp)
{
	MOD_INC_USE_COUNT;
	return 0;
}

static int dev_perfctr_release(struct inode *inode, struct file *filp)
{
	MOD_DEC_USE_COUNT;
	return 0;
}

static struct file_operations dev_perfctr_file_ops = {
	.ioctl = dev_perfctr_ioctl,
	.open = dev_perfctr_open,
	.release = dev_perfctr_release,
};

static struct miscdevice dev_perfctr = {
	.minor = 182,
	.name = "perfctr",
	.fops = &dev_perfctr_file_ops,
};

#define STR2(X) #X
#define STR(X)	STR2(X)

int __init perfctr_init(void)
{
	int err = perfctr_cpu_init();
	if( err ) {
		printk(KERN_INFO "perfctr: not supported by this processor\n");
		return err;
	}
	if( (err = misc_register(&dev_perfctr)) != 0 ) {
		printk(KERN_ERR "/dev/perfctr: failed to register, errno %d\n",
		       -err);
		return -ENODEV;
	}
	vperfctr_init();
	gperfctr_init();
	perfctr_info.version_major = PERFCTR_VERSION_MAJOR;
	perfctr_info.version_minor = PERFCTR_VERSION_MINOR;
	perfctr_info.version_micro = PERFCTR_VERSION_MICRO;
	perfctr_info.cpu_type = perfctr_cpu_type;
	perfctr_info.nrcpus = smp_num_cpus;
	printk(KERN_INFO "perfctr: driver version %u.%u%s\n",
	       PERFCTR_VERSION_MAJOR, PERFCTR_VERSION_MINOR,
	       PERFCTR_VERSION_MICRO ? "." STR(PERFCTR_VERSION_MICRO) : "");
	return 0;
}

#ifdef MODULE
int init_module(void)
{
	return perfctr_init();
}

void cleanup_module(void)
{
	vperfctr_cleanup();
	misc_deregister(&dev_perfctr);
}
#endif
