/*
 * x86 performance-monitoring counters driver.
 *
 * Copyright (C) 1999-2000  Mikael Pettersson
 */

/* Initialise the hardware driver. */
extern int perfctr_cpu_init(void);

/* Reserve the hardware for a particular service. */
extern const char *perfctr_cpu_reserve(const char *service);
extern void perfctr_cpu_release(const char *service);

/* Save current TSC/PMC counter values in *ctrs. */
extern void (*perfctr_cpu_read_counters)(int nrctrs, struct perfctr_ctrs *ctrs);

/* Write control data to the CPU's perfctr evntsel MSRs.
   This is to be done lazily: if a given evntsel MSR already
   contains the correct value, do not write to it. */
extern void (*perfctr_cpu_write_control)(int nrctrs, const struct perfctr_control *control);

/* Return -EPERM if the control info is invalid, otherwise the
   highest numbered enabled counter (starting from 1). */
extern int (*perfctr_cpu_check_control)(const struct perfctr_control *control);

/* Detected CPU type. */
extern unsigned char perfctr_cpu_type;

/* Toggle user-mode read-access to the counter registers. */
extern int perfctr_cpu_have_rdpmc;

static __inline__ unsigned int get_cr4(void)
{
	unsigned int cr4;
	__asm__ __volatile__("movl %%cr4,%0" : "=r"(cr4));
	return cr4;
}

static __inline__ void set_cr4(unsigned int cr4)
{
	__asm__ __volatile__("movl %0,%%cr4" : : "r"(cr4));
}

static __inline__ void perfctr_cpu_enable_rdpmc(void)
{
	if( perfctr_cpu_have_rdpmc )
		set_cr4(get_cr4() | X86_CR4_PCE);
}

static __inline__ void perfctr_cpu_disable_rdpmc(void)
{
	if( perfctr_cpu_have_rdpmc )
		set_cr4(get_cr4() & ~X86_CR4_PCE);
}
